import os
import time

from fileio import File


_tfdsio_dir = os.path.abspath(os.path.dirname(__file__))
_tfdsio_imports = os.path.join(_tfdsio_dir, 'imported')

def set_adc(path):
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = path

def timer(start_time=None):
    if not start_time:
        return time.time()
    return time.time() - start_time



import os
import sys
import numpy as np
import tensorflow as tf
import posixpath
import termcolor
from dataclasses import dataclass, asdict


from absl import logging
import tensorflow_datasets as tfds
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core.utils import Version, PathLike, version


from tfdsio.utils import File, _tfdsio_dir, _tfdsio_imports
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Union

_feats = {
    'text': tfds.features.Text(),
    'audio': tfds.features.Audio(),
    'image': tfds.features.Image(),
}

_splits = {
    'train': tfds.Split.TRAIN,
    'validation': tfds.Split.VALIDATION,
    'test': tfds.Split.TEST
}

_parsers = {
    'jsonlines': File.tfjl,
    'text': File.tftl,
    'csv': File.tfcsv,
}


VersionOrStr = Union[Version, str]
_has_logged_ds_info = set()

def log_dataset_info(name, msg):
    global _has_logged_ds_info
    if name in _has_logged_ds_info:
        return
    termcolor.cprint(msg, attrs=["bold"])
    _has_logged_ds_info.add(name)
    

def create_features(features):
    return tfds.features.FeaturesDict({key: _feats[val] for key, val in features.items()})


@dataclass
class BuilderConfig:
    name: Optional[str] = None
    classifier: Optional[str] = None
    version: Optional[VersionOrStr] = '1.0.0'
    release_notes: Optional[Dict[str, str]] = None
    supported_versions: List[str] = None
    description: Optional[str] = ''
    dataset_urls: Optional[Any] = None
    dataset_format: Optional[str] = 'jsonlines'
    features: Optional[Any] = None
    datamap: Optional[Any] = None
    supervised_keys: Optional[Any] = None
    homepage: Optional[str] = ''
    citation: Optional[Any] = None
    metadata: Optional[Any] = None
    redistribution_info: Optional[Any] = None
    data_dir: Optional[str] = None
    process_args: Optional[Any] = None

class TFDSIOConfig(BuilderConfig):
    def parse_config(self):
        for key, val in self.__dict__.items():
            if key == 'features' and val:
                self.__dict__[key] = create_features(val)
            elif key == 'supported_versions' and not val:
                self.__dict__[key] = []
            elif key == 'dataset_format':
                assert val in list(_parsers.keys())
            elif key == 'dataset_urls' and val:
                _dsrc = {'extra': []}
                if isinstance(val, str):
                    _dsrc['train'] = val
                    
                elif isinstance(val, dict):
                    for k,v in val.items():
                        if k in ['train', 'training']:
                            _dsrc['train'] = v
                        elif k in ['dev', 'develop', 'development', 'val', 'validation']:
                            _dsrc['validation'] = v
                        elif k in ['test', 'gold']:
                            _dsrc['test'] = v
                        else:
                            if isinstance(v, str):
                                _dsrc['extra'].append(v)
                            elif isinstance(v, list):
                                _dsrc['extra'].extend(v)
                            else:
                                _dsrc['extra'] = v

                self.__dict__[key] = _dsrc.copy()
            
            else:
                self.__dict__[key] = val


    def from_config(self, config):
        for k,v in config.items():
            if k in self.__dict__:
                self.__dict__[k] = v
        self.parse_config()

    def from_file(self, filename):
        assert File.exists(filename)
        config = File.jsonload(filename)
        self.from_config(config)
    
    def from_auto(self, config_or_file):
        if isinstance(config_or_file, dict):
            return self.from_config(config_or_file)
        return self.from_file(config_or_file)
    
    @property
    def dataset_info(self):
        keys = ['description', 'features', 'supervised_keys', 'homepage', 'citation', 'metadata', 'redistribution_info']
        _info = {k: self.__dict__[k] for k in keys}
        return _info
    
    @property
    def dataset_name(self):
        return f'{self.name}/{self.classifier}:{self.version}'

class TFDSIOInfo(tfds.core.DatasetInfo):
    @property
    def full_name(self):
        """Full canonical name: (<dataset_name>/<dataset_classifier>/<version>)."""
        names = []
        if self._builder.builder_config:
            names.append(self._builder.builder_config.name)
            names.append(self._builder.builder_config.classifier)
            names.append(str(self.version))
        return posixpath.join(*names)

class TFDSIOCorpus(tfds.core.GeneratorBasedBuilder):
    VERSION = None
    RELEASE_NOTES: ClassVar[Dict[str, str]] = {}
    SUPPORTED_VERSIONS = []
    BUILDER_CONFIGS = []
    MANUAL_DOWNLOAD_INSTRUCTIONS = None
    def __init__(self, *, config: Union[None, str, BuilderConfig] = None, **kwargs: Any):
        if config.data_dir:
            config.data_dir = os.fspath(config.data_dir)  # Pathlib -> str
        elif os.environ.get('TFDS_DIR', None):
            config.data_dir = os.fspath(os.environ['TFDS_DIR'])
        elif os.environ.get('GCS_DIR', None):
            config.data_dir = os.fspath(os.environ['GCS_DIR'])

        file_format = kwargs.get('file_format', file_adapters.DEFAULT_FILE_FORMAT)
        self._original_state = dict(data_dir=config.data_dir, config=config, version=config.version)
        self._builder_config = self._create_builder_config(config)
        self.__dict__['name'] = self._builder_config.name
        self._version = self._pick_version(config.version)
        self._data_dir_root, self._data_dir = self._build_data_dir(config)
        if tf.io.gfile.exists(self._data_dir):
            self.info.read_from_directory(self._data_dir)
            self._log_dataset_info()
        else:
            self.info.initialize_from_bucket()
        
        try:
            self._file_format = file_adapters.FileFormat(file_format)
        except ValueError:
            all_values = [f.value for f in file_adapters.FileFormat]
            raise ValueError(f"{file_format} is not a valid format. Valid file formats: {all_values}")

    
    def _log_dataset_info(self):
        msg = "--------------------------------------------\n"
        msg += f"TFDSIO Dataset: {self.info.full_name}\n"
        msg += "Splits:\n"
        for split, details in self.info.splits.items():
            msg += f" - {split}: {details.num_examples} examples, {details.num_shards} shards\n"
        msg += f"Features: {str(self.info.features)}\n"
        msg += f"Download Size: {self.info.download_size}, Dataset Size: {self.info.dataset_size}\n"
        msg += "--------------------------------------------"
        log_dataset_info(self.info.full_name, msg)
    
    def _log_download_done(self):
        msg = f"Dataset {self.info.full_name} downloaded and prepared to {self._data_dir}.\n"
        termcolor.cprint(msg, attrs=["bold"])
        self._log_dataset_info()
    
    def _log_download_bytes(self):
        termcolor.cprint(
            f"Downloading and Preparing dataset {self.info.full_name} - {self.info.download_size} "
            f"(download: {self.info.download_size}, "
            f"generated: {self.info.dataset_size}, "
            f"total: {self.info.download_size + self.info.dataset_size}) "
            f"\nto {self._data_dir}...",
            attrs=["bold"],
        )
    
    def _split_generators(self, dl_manager):
        dataset_urls = self.builder_config.dataset_urls.copy()
        _extra = dataset_urls.pop('extra', None)
        dl_paths = dl_manager.download(dataset_urls)
        msg = f"Downloading Temp Files to:\n"
        for fs, fn in dl_paths.items():
            msg += f" - {fs}: {str(fn)}\n"
        termcolor.cprint(msg)
        if _extra:
            dl_manager.download(_extra)
        dataset_splits = list(dataset_urls.keys())
        file_splits = [tfds.core.SplitGenerator(name=_splits[split], gen_kwargs={"paths": dl_paths[split], "extra": _extra}) for split in dataset_splits]
        return file_splits
    
    def _generate_examples(self, paths, extra=None, **kwargs):
        pipeline = _parsers[self.builder_config.dataset_format](paths)
        _do_preprocess = self.has_preprocessor
        idx = 0
        for n, data in enumerate(pipeline):
            if _do_preprocess:
                data = self.preprocessor(idx=idx, data=data, extra=extra, **kwargs)
            if data:
                if isinstance(data, dict):
                    if self.builder_config.datamap:
                        data = {v: data.get(k, '') for k,v in self.builder_config.datamap.items()}
                    yield idx, data
                    idx += 1
                
                elif isinstance(data, (list, np.array)) and _do_preprocess:
                    for ex in data:
                        if self.builder_config.datamap:
                            ex = {v: ex.get(k, '') for k,v in self.builder_config.datamap.items()}
                        yield idx, ex
                        idx += 1

                elif isinstance(data, (list, np.array)):
                    data = {self.builder_config.features[x]: data[x] for x in enumerate(data)}
                    yield idx, data
                    idx += 1
                
                elif isinstance(data, (int, float, str, bytes)):
                    data = {self.builder_config.features[0]: data}
                    yield idx, data
                    idx += 1

    def _relative_data_dir(self, with_version=True):
        """Relative path of this dataset in data_dir."""
        builder_data_dir = self.name
        builder_config = self._builder_config
        if builder_config:
            if builder_config.classifier:
                builder_data_dir = os.path.join(builder_data_dir, builder_config.classifier)
            else:
                builder_data_dir = os.path.join(builder_config.name)
        
        if not with_version:
            return builder_data_dir
        version_data_dir = os.path.join(builder_data_dir, str(self._version))
        return version_data_dir

    def _build_data_dir(self, config):
        """Return the data directory for the current version.
        Args:
        given_data_dir: `Optional[str]`, root `data_dir` passed as
            `__init__` argument.
        Returns:
        data_dir_root: `str`, The root dir containing all datasets, downloads,...
        data_dir: `str`, The version data_dir
            (e.g. `<data_dir_root>/<ds_name>/<config>/<version>`)
        """
        given_data_dir = config.data_dir
        builder_dir = self._relative_data_dir(with_version=False)
        version_dir = self._relative_data_dir(with_version=True)
        default_data_dir = constants.get_default_data_dir(given_data_dir=given_data_dir)
        all_data_dirs = constants.list_data_dirs(given_data_dir=given_data_dir)
        all_versions = set()
        requested_version_dirs = {}
        for data_dir_root in all_data_dirs:
            full_builder_dir = os.path.join(data_dir_root, builder_dir)
            data_dir_versions = set(version.list_all_versions(full_builder_dir))
            if self.version in data_dir_versions:
                requested_version_dirs[data_dir_root] = os.path.join(data_dir_root, version_dir)
            all_versions.update(data_dir_versions)

        if len(requested_version_dirs) > 1:
            raise ValueError("Dataset was found in more than one directory: {}. Please resolve the ambiguity by explicitly specifying `data_dir=`.".format(requested_version_dirs.values()))
        elif len(requested_version_dirs) == 1:  # The dataset is found once
            return next(iter(requested_version_dirs.items()))

        # No dataset found, use default directory
        data_dir = os.path.join(default_data_dir, version_dir)
        if all_versions:
            logging.warning(
                "Found a different version of the requested dataset:\n"
                "%s\n"
                "Using %s instead.",
                "\n".join(str(v) for v in sorted(all_versions)),
                data_dir
            )
        return default_data_dir, data_dir

    def set_preprocessor(self, function):
        self._proc = function
    
    @property
    def has_preprocessor(self):
        return bool('_proc' in self.__dict__)

    def preprocessor(self, *args, **kwargs):
        if self.has_preprocessor:
            if self.builder_config.process_args:
                kwargs.update(self.builder_config.process_args)
            return self._proc(*args, **kwargs)
        return None

    @property
    def libname(self):
        return 'tfdsio'

    def _info(self) -> tfds.core.DatasetInfo:
        return TFDSIOInfo(builder=self, **self.builder_config.dataset_info)


def tfds_dataset(config_or_file, preprocessor=None, build=True, as_tfdataset=True, as_numpy=False, as_df=False, **kwargs):
    config = TFDSIOConfig()
    config.from_auto(config_or_file)
    builder = TFDSIOCorpus(config=config, **kwargs)
    #builder = TFDSIOCorpus(data_dir=config.data_dir, config=config, version=config.version, **kwargs)
    if preprocessor:
        builder.set_preprocessor(preprocessor)
    if build:
        build_args = kwargs.get('build_args', {})
        builder.download_and_prepare(**build_args)
    if as_tfdataset or as_numpy:
        dataset_args = kwargs.get('dataset_args', {})
        ds = builder.as_dataset(**dataset_args)
        if as_numpy:
            return tfds.as_numpy(ds)
        if as_df:
            return tfds.as_dataframe(ds, builder.info)
        return ds
    return builder

def tfds_remove(config_or_file, prompt=True):
    config = TFDSIOConfig()
    config.from_auto(config_or_file)
    builder = TFDSIOCorpus(config=config)
    del_dir = os.fspath(builder.__cached_info.data_path)
    full_name = builder.__cached_info.full_name
    if prompt:
        del_conf = input(f'Deleting TFDSIO {full_name} at {del_dir} - Confirm [y/n]')
        if del_conf.lower() != 'y':
            logging.info('Exiting')
            return
    File.rmdir(del_dir)
    logging.info(f'Removed: {del_dir}')


def tfds_sample(dataset, num_samples=5, return_samples=True):
    samples = []
    for ex in tfds.as_numpy(dataset.take(num_samples)):
        logging.info(ex)
        samples.append(ex)
    if return_samples:
        return samples




# Adapted from https://github.com/rwightman/pytorch-image-models/blob/benchmark-fixes-vit_hybrids/timm/data/parsers/parser_tfds.py

import os
import io
import math
import numpy as np
import tensorflow as tf
from tfdsio.dataset import tfds
from tfdsio.dataset import TFDSIOConfig, TFDSIOCorpus

try:
    import torch
    import torch.distributed as dist
    _torch_installed = True
except ImportError as e:
    _torch_installed = False

try:
    import transformers
    from transformers import AutoTokenizer
    _transformers_installed = True
except ImportError as e:
    _transformers_installed = False

try:
    from PIL import Image
    _pil_installed = True
except ImportError as e:
    _pil_installed = False


def assert_torch():
    if _torch_installed:
        return
    print("Please install torch package `pip install torch`.")
    exit(1)

def assert_transformers():
    if _transformers_installed:
        os.environ['TOKENIZERS_PARALLELISM'] = 'true'
        return
    print("Please install transformers package `pip install transformers`.")
    exit(1)


def assert_pil():
    if _pil_installed:
        return
    print("Please install PIL package `pip install PIL`.")
    exit(1)


def even_split_indices(split, n, num_samples):
    partitions = [round(i * num_samples / n) for i in range(n + 1)]
    return [f"{split}[{partitions[i]}:{partitions[i+1]}]" for i in range(n)]


class TorchTFDSIO(object):
    """ Wrap Tensorflow Datasets for use in PyTorch
    There several things to be aware of:
      * To prevent excessive samples being dropped per epoch w/ distributed training or multiplicity of
         dataloader workers, the train iterator wraps to avoid returning partial batches that trigger drop_last
         https://github.com/pytorch/pytorch/issues/33413
      * With PyTorch IterableDatasets, each worker in each replica operates in isolation, the final batch
        from each worker could be a different size. For training this is worked around by option above, for
        validation extra samples are inserted iff distributed mode is enabled so that the batches being reduced
        across replicas are of same size. This will slightly alter the results, distributed validation will not be
        100% correct. This is similar to common handling in DistributedSampler for normal Datasets but a bit worse
        since there are up to N * J extra samples with IterableDatasets.
      * The sharding (splitting of dataset into TFRecord) files imposes limitations on the number of
        replicas and dataloader workers you can use. For really small datasets that only contain a few shards
        you may have to train non-distributed w/ 1-2 dataloader workers. This is likely not a huge concern as the
        benefit of distributed training or fast dataloading should be much less for small datasets.
      * This wrapper is currently configured to return individual, decompressed image samples from the TFDS
        dataset. The augmentation (transforms) and batching is still done in PyTorch. It would be possible
        to specify TF augmentation fn and return augmented batches w/ some modifications to other downstream
        components.
    """
    def __init__(self, config_or_file, preprocessor=None, tokenizer=None, split='train', data_type='text', shuffle=False, is_training=False, batch_size=None, repeats=0, seed=42, shuffle_size=16834, prefetch_size=4096, max_tf_threadpool=8, **kwargs):
        super().__init__()
        assert_torch()
        self.tokenizer = None
        if tokenizer:
            assert_transformers()
            if isinstance(tokenizer, str):
                self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
            else:
                self.tokenizer = tokenizer
            self.tokenizer_args = kwargs.get('tokenizer_args', {})
        self.config = TFDSIOConfig()
        self.config.from_auto(config_or_file)
        self.datatype = data_type
        self.split = split
        self.seed, self.shuffle_size, self.prefetch_size, self.max_tf_threadpool = seed, shuffle_size, prefetch_size, max_tf_threadpool
        self.shuffle = shuffle
        self.is_training = is_training
        if self.is_training:
            assert batch_size is not None,\
                "Must specify batch_size in training mode for reasonable behaviour w/ TFDS wrapper"
        self.batch_size = batch_size
        self.repeats = repeats
        self.subsplit = None
        self.builder = TFDSIOCorpus(data_dir=self.config.data_dir, config=self.config, version=self.config.version)
        if self.preprocessor:
            self.builder.set_preprocessor(preprocessor)
        
        self.builder.download_and_prepare()
        self.num_samples = self.builder.info.splits[split].num_examples
        self.ds = None
        self.worker_info = None
        self.dist_rank = 0
        self.dist_num_replicas = 1
        if dist.is_available() and dist.is_initialized() and dist.get_world_size() > 1:
            self.dist_rank = dist.get_rank()
            self.dist_num_replicas = dist.get_world_size()

    def _lazy_init(self):
        """ Lazily initialize the dataset.
        This is necessary to init the Tensorflow dataset pipeline in the (dataloader) process that
        will be using the dataset instance. The __init__ method is called on the main process,
        this will be called in a dataloader worker process.
        NOTE: There will be problems if you try to re-use this dataset across different loader/worker
        instances once it has been initialized. Do not call any dataset methods that can call _lazy_init
        before it is passed to dataloader.
        """
        worker_info = torch.utils.data.get_worker_info()
        split = self.split
        num_workers = 1
        if worker_info is not None:
            self.worker_info = worker_info
            num_workers = worker_info.num_workers
            global_num_workers = self.dist_num_replicas * num_workers
            worker_id = worker_info.id
            if not self.is_training and '[' not in self.split:
                self.subsplit = even_split_indices(self.split, global_num_workers, self.num_samples)[self.dist_rank * num_workers + worker_id]

        if self.subsplit is None:
            input_context = tf.distribute.InputContext(num_input_pipelines=self.dist_num_replicas * num_workers, input_pipeline_id=self.dist_rank * num_workers + worker_id,
                num_replicas_in_sync=self.dist_num_replicas
            )
        else:
            input_context = None

        read_config = tfds.ReadConfig(shuffle_seed=self.seed, shuffle_reshuffle_each_iteration=True, input_context=input_context)
        ds = self.builder.as_dataset(split=self.subsplit or self.split, shuffle_files=self.shuffle, read_config=read_config)
        ds.options().experimental_threading.private_threadpool_size = max(1, self.max_tf_threadpool // num_workers)
        ds.options().experimental_threading.max_intra_op_parallelism = 1
        if self.is_training or self.repeats > 1:
            ds = ds.repeat()
        if self.shuffle:
            ds = ds.shuffle(min(self.num_samples // self._num_pipelines, self.shuffle_size), seed=0)
        ds = ds.prefetch(min(self.num_samples // self._num_pipelines, self.prefetch_size))
        self.ds = tfds.as_numpy(ds)

    def __iter__(self):
        if self.ds is None:
            self._lazy_init()
        if self.datatype in ['image', 'img']:
            assert_pil()
            return self._img_iter()
        return self._text_iter()


    def _text_iter(self):
        target_sample_count = math.ceil(max(1, self.repeats) * self.num_samples / self._num_pipelines)
        if self.is_training:
            target_sample_count = math.ceil(target_sample_count / self.batch_size) * self.batch_size
        idx = 0
        for example in self.ds:
            for k,v in example.items():
                if isinstance(v, np.ndarray):
                    pass
                
                else:
                    v_text = tf.compat.as_text(v)
                    if self.tokenizer:
                        example[k] = {'text': v_text, 'tokens': self.tokenizer(v_text, **self.tokenizer_args)}
                    else:
                        example[k] = v_text
            yield example
            idx += 1
            if self.is_training and idx >= target_sample_count:
                break
        if not self.is_training and self.dist_num_replicas and 0 < idx < target_sample_count:
            yield example
            idx += 1


    def _img_iter(self):
        # compute a rounded up sample count that is used to:
        #   1. make batches even cross workers & replicas in distributed validation.
        #     This adds extra samples and will slightly alter validation results.
        #   2. determine loop ending condition in training w/ repeat enabled so that only full batch_size
        #     batches are produced (underlying tfds iter wraps around)
        target_sample_count = math.ceil(max(1, self.repeats) * self.num_samples / self._num_pipelines)
        if self.is_training:
            # round up to nearest batch_size per worker-replica
            target_sample_count = math.ceil(target_sample_count / self.batch_size) * self.batch_size
        sample_count = 0
        for example in self.ds:
            for k,v in example.items():
                example[k] = tf.compat.as_text(v)
            img = Image.fromarray(sample['image'], mode='RGB')
            yield img, sample['label']
            sample_count += 1
            if self.is_training and sample_count >= target_sample_count:
                # Need to break out of loop when repeat() is enabled for training w/ oversampling
                # this results in extra samples per epoch but seems more desirable than dropping
                # up to N*J batches per epoch (where N = num distributed processes, and J = num worker processes)
                break
        if not self.is_training and self.dist_num_replicas and 0 < sample_count < target_sample_count:
            # Validation batch padding only done for distributed training where results are reduced across nodes.
            # For single process case, it won't matter if workers return different batch sizes.
            # FIXME if using input_context or % based subsplits, sample count can vary by more than +/- 1 and this
            # approach is not optimal
            yield img, sample['label']  # yield prev sample again
            sample_count += 1


    @property
    def _num_workers(self):
        return 1 if self.worker_info is None else self.worker_info.num_workers

    @property
    def _num_pipelines(self):
        return self._num_workers * self.dist_num_replicas

    def __len__(self):
        # this is just an estimate and does not factor in extra samples added to pad batches based on
        # complete worker & replica info (not available until init in dataloader).
        return math.ceil(max(1, self.repeats) * self.num_samples / self.dist_num_replicas)

    def _filename(self, index, basename=False, absolute=False):
        assert False, "Not supported" # no random access to samples

    def filenames(self, basename=False, absolute=False):
        """ Return all filenames in dataset, overrides base"""
        if self.ds is None:
            self._lazy_init()
        names = []
        for sample in self.ds:
            if len(names) > self.num_samples:
                break  # safety for ds.repeat() case
            if 'file_name' in sample:
                name = sample['file_name']
            elif 'filename' in sample:
                name = sample['filename']
            elif 'id' in sample:
                name = sample['id']
            else:
                assert False, "No supported name field present"
            names.append(name)
        return names




from .adapter import (
    T5DataSource,
    load_sp_model,
    create_sp_model,
    set_vocab,
    default_output_features,
    default_preprocessor,
    default_metrics,
    add_to_registry,
    remove_from_registry,
    add_to_mixture,
    remove_from_mixture,
    get_registry,
    get_mixtures,
    sample_registry,
    sample_mixture
)

from .adapter import t5_preprocessors as preprocessors
from fileio import TFDSIODataset as DatasetConfig



import os
import sys
import os.path

from setuptools.command.install import install as _install
from setuptools import setup, find_packages

root = os.path.abspath(os.path.dirname(__file__))
package_name = "tfdsio"
packages = find_packages(
    include=[package_name, "{}.*".format(package_name)]
)

__version_info__ = (0, 0, 6)
version = ".".join(map(str, __version_info__))
binary_names = [package_name]

with open(os.path.join(root, 'README.md'), 'rb') as readme:
    long_description = readme.read().decode('utf-8')

setup(
    name=package_name,
    version=version,
    description="tfdsio",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    url='http://github.com/trisongz/tfdsio',
    python_requires='>=3.6',
    install_requires=[
        "tensorflow>=1.15.0",
        "tensorflow_datasets>=4.2.0",
        "file-io>=0.1.0",
        "seqio",
        "t5",
        "sentencepiece",
    ],
    packages=packages,
    extras_require={
        'transformers': ['transformers', 'sentencepiece'],
        't5': ['t5', 'seqio'],
    },
    entry_points={},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
    ],
    data_files=[],
    include_package_data=True
)



# adapter for T5 from https://github.com/google-research/text-to-text-transfer-transformer/blob/master/t5/seqio/utils.py

import abc
import collections
import inspect
import json
import os
import re
import functools

from absl import logging
import dataclasses
import numpy as np
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds
from typing import Any, Callable, Iterable, Mapping, MutableMapping, Optional, Sequence, Tuple, Type, Union
from tfdsio.dataset import TFDSIOConfig, TFDSIOCorpus

#try:
import t5
import seqio
from pprint import pprint
from fileio import File
from sentencepiece import sentencepiece_model_pb2 as model
from tensorflow.python.compat.v2_compat import enable_v2_behavior

_vocab_path = None
_num_extra_ids = 100
_default_sp_models = {
    'default': {
        'path': 'gs://t5-data/vocabs/cc_all.32000/sentencepiece.model',
        'size': 32000
    },
    'multi': {
        'path': 'gs://t5-data/vocabs/mc4.250000.100extra/sentencepiece.model',
        'size': 250100
    }
}

#_dataset_mixtures = {}
_dataset_registry = []
_dataset_mixtures = {}

_TFDS_DATA_DIR_OVERRIDE = seqio.utils._TFDS_DATA_DIR_OVERRIDE
_GLOBAL_CACHE_DIRECTORIES = seqio.utils._GLOBAL_CACHE_DIRECTORIES
Vocabulary = seqio.vocabularies.Vocabulary

#_seqio_enabled = True
#except ImportError:
#    _TFDS_DATA_DIR_OVERRIDE = None
#    _GLOBAL_CACHE_DIRECTORIES = []
#    Vocabulary = Any
#    _seqio_enabled = False
#
#try:
#    import t5
#    _t5_enabled = True
#except ImportError:
#    t5 = None
#    _t5_enabled = False

@dataclasses.dataclass(frozen=True)
class Feature:
    """A container for attributes of output features of data providers."""
    vocabulary: Vocabulary
    add_eos: bool = True
    required: bool = True
    dtype: tf.DType = tf.int32


@dataclasses.dataclass(frozen=True)
class ShardInfo:
    """A container for specifying sharding info."""
    index: int
    num_shards: int


class DatasetProviderBase(metaclass=abc.ABCMeta):
    """Abstract base for classes that provide a tf.data.Dataset."""

    @abc.abstractproperty
    def output_features(self) -> Mapping[str, Feature]:
        raise NotImplementedError

    @abc.abstractproperty
    def splits(self) -> Sequence[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_dataset(self, sequence_length: int, split: str, use_cached: bool = False, shuffle: bool = True,
        seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None, num_epochs: int = 1) -> tf.data.Dataset:
        """Returns the requested tf.data.Dataset."""
        raise NotImplementedError

    @abc.abstractmethod
    def num_input_examples(self, split: str) -> int:
        raise NotImplementedError


class DatasetProviderRegistry(object):
    _REGISTRY: MutableMapping[str, DatasetProviderBase]
    _PROVIDER_TYPE: Type[DatasetProviderBase]

    @classmethod
    def add_provider(cls, name: str, provider):
        """Adds a data provider instance to the registry."""
        if name in cls._REGISTRY:
            raise ValueError("Attempting to register duplicate provider: %s" % name)
        if not isinstance(provider, cls._PROVIDER_TYPE):
            raise ValueError("Attempting to register a class not of an invalid type. Expecting instance of %s, got %s" % (cls._PROVIDER_TYPE, type(provider).__name__))
        cls._REGISTRY[name] = provider

    @classmethod
    def add(cls, name: str, provider_cls, *provider_args, **provider_kwargs):
        """Instantiates and adds provider to the registry."""
        if not issubclass(provider_cls, cls._PROVIDER_TYPE):
            raise ValueError("Attempting to register a class not of an invalid type. Expecting instance of %s, got %s" %(cls._PROVIDER_TYPE, provider_cls))
        provider = provider_cls(*provider_args, **provider_kwargs)
        cls.add_provider(name, provider)
        return provider

    @classmethod
    def remove(cls, name):
        """Remove provider from the registry, if it exists."""
        if name in cls._REGISTRY:
            del cls._REGISTRY[name]

    @classmethod
    def get(cls, name):
        """Returns provider from the registry."""
        if name not in cls._REGISTRY:
            raise ValueError("Provider name not registered: %s" % name)
        return cls._REGISTRY[name]

    @classmethod
    def names(cls):
        """Returns all provider names in registry."""
        return cls._REGISTRY.keys()

    @classmethod
    def reset(cls):
        """Removes all of the registered tasks."""
        cls._REGISTRY = {}

    @classmethod
    def get_dataset(cls, name, sequence_length, split, use_cached=False, shuffle=True, seed=None, shard_info=None, num_epochs=1):
        """Returns the requested tf.data.Dataset."""
        return cls.get(name).get_dataset(sequence_length=sequence_length, split=split, use_cached=use_cached, shuffle=shuffle, seed=seed, shard_info=shard_info, num_epochs=num_epochs)



class LazyTFDSIOLoader(object):
    """Wrapper for TFDS datasets with memoization and additional functionality.
    Lazily loads info from TFDS and provides memoization to avoid expensive hidden
    file operations. Also provides additional utility methods.
    """

    _MEMOIZED_BUILDERS = {}

    def __init__(self, config_or_file, preprocessor=None, split_map=None):
        """LazyTfdsLoader constructor.
        Args:
        name: str, the name of the TFDS dataset.
        data_dir: str (optional), directory to read/write TFDS data.
        split_map: dict (optional), mapping from canonical splits
            (e.g., 'validation') to TFDS splits or slices
            (e.g., 'train[':1%']).
        """
        self._config = TFDSIOConfig()
        self._config.from_auto(config_or_file)
        self._preprocessor = preprocessor

        self._name = self._config.dataset_name
        self._data_dir = self._config.data_dir
        self._split_map = split_map

    @property
    def name(self):
        return self._name

    @property
    def data_dir(self):
        if _TFDS_DATA_DIR_OVERRIDE:
            if self._data_dir:
                logging.warning("Overriding TFDS data directory '%s' with '%s' for dataset '%s'.", self._data_dir, _TFDS_DATA_DIR_OVERRIDE, self.name)
            return _TFDS_DATA_DIR_OVERRIDE
        return self._data_dir

    @property
    def builder(self):
        builder_key = (self.name, self.data_dir)
        if builder_key not in LazyTFDSIOLoader._MEMOIZED_BUILDERS:
            LazyTFDSIOLoader._MEMOIZED_BUILDERS[builder_key] = TFDSIOCorpus(config=self._config)
        return LazyTFDSIOLoader._MEMOIZED_BUILDERS[builder_key]

    @property
    def info(self):
        return self.builder.info

    def _map_split(self, split):
        return self._split_map[split] if self._split_map else split

    def files(self, split):
        """Returns set of instructions for reading TFDS files for the dataset."""
        split = self._map_split(split)

        if "/" not in self.name and self.builder.BUILDER_CONFIGS:
            # If builder has multiple configs, and no particular config was
            # requested, raise an error.
            raise ValueError("Dataset '%s' has multiple configs." % self.name)

        split_info = self.builder.info.splits[split]
        files = split_info.file_instructions

        if not files:
            logging.fatal("No TFRecord files found for dataset: %s", self.name)
        return files

    def load(self, split, shuffle_files, seed=None, shard_info=None):
        """Returns a tf.data.Dataset for the given split."""
        split = self._map_split(split)
        input_context = (tf.distribute.InputContext(num_input_pipelines=shard_info.num_shards, input_pipeline_id=shard_info.index) if shard_info else None)
        builder = TFDSIOCorpus(config=self._config)
        if self._preprocessor:
            builder.set_preprocessor(self._preprocessor)
        builder.download_and_prepare()
        ds = builder.as_dataset(split=split, shuffle_files=shuffle_files, read_config=tfds.ReadConfig(shuffle_seed=seed, skip_prefetch=True, input_context=input_context))
        return ds

    def load_shard(self, file_instruction, shuffle_files=False, seed=None):
        """Returns a dataset for a single shard of the TFDS TFRecord files."""
        ds = self.builder._tfrecords_reader.read_files([file_instruction], read_config=tfds.ReadConfig(shuffle_seed=seed), shuffle_files=shuffle_files)
        return ds

    def size(self, split):
        """Returns the number of examples in the split."""
        split = self._map_split(split)
        ds_splits = self.info.splits
        dataset_size = ds_splits[split].num_examples
        # Very large datasets have num_examples = 0; default instead to np.inf
        dataset_size = dataset_size if dataset_size > 0 else np.inf
        return dataset_size


class LazyTfdsLoader(object):
    """Wrapper for TFDS datasets with memoization and additional functionality.
    Lazily loads info from TFDS and provides memoization to avoid expensive hidden
    file operations. Also provides additional utility methods.
    """

    _MEMOIZED_BUILDERS = {}

    def __init__(self, name, data_dir=None, split_map=None):
        """LazyTfdsLoader constructor.
        Args:
        name: str, the name of the TFDS dataset.
        data_dir: str (optional), directory to read/write TFDS data.
        split_map: dict (optional), mapping from canonical splits
            (e.g., 'validation') to TFDS splits or slices
            (e.g., 'train[':1%']).
        """
        self._name = name
        self._data_dir = data_dir
        self._split_map = split_map

    @property
    def name(self):
        return self._name

    @property
    def data_dir(self):
        if _TFDS_DATA_DIR_OVERRIDE:
            if self._data_dir:
                logging.warning("Overriding TFDS data directory '%s' with '%s' for dataset '%s'.", self._data_dir, _TFDS_DATA_DIR_OVERRIDE, self.name)
            return _TFDS_DATA_DIR_OVERRIDE
        return self._data_dir

    @property
    def builder(self):
        builder_key = (self.name, self.data_dir)
        if builder_key not in LazyTfdsLoader._MEMOIZED_BUILDERS:
            LazyTfdsLoader._MEMOIZED_BUILDERS[builder_key] = tfds.builder(self.name, data_dir=self.data_dir)
        return LazyTfdsLoader._MEMOIZED_BUILDERS[builder_key]

    @property
    def info(self):
        return self.builder.info

    def _map_split(self, split):
        return self._split_map[split] if self._split_map else split

    def files(self, split):
        """Returns set of instructions for reading TFDS files for the dataset."""
        split = self._map_split(split)

        if "/" not in self.name and self.builder.BUILDER_CONFIGS:
            # If builder has multiple configs, and no particular config was
            # requested, raise an error.
            raise ValueError("Dataset '%s' has multiple configs." % self.name)

        split_info = self.builder.info.splits[split]
        files = split_info.file_instructions

        if not files:
            logging.fatal("No TFRecord files found for dataset: %s", self.name)
        return files

    def load(self, split, shuffle_files, seed=None, shard_info=None):
        """Returns a tf.data.Dataset for the given split."""
        split = self._map_split(split)
        input_context = (tf.distribute.InputContext(num_input_pipelines=shard_info.num_shards, input_pipeline_id=shard_info.index) if shard_info else None)
        return tfds.load(self._name,
            split=split,
            data_dir=self.data_dir,
            shuffle_files=shuffle_files,
            download=True,
            try_gcs=True,
            read_config=tfds.ReadConfig(shuffle_seed=seed, skip_prefetch=True, input_context=input_context)
        )

    def load_shard(self, file_instruction, shuffle_files=False, seed=None):
        """Returns a dataset for a single shard of the TFDS TFRecord files."""
        ds = self.builder._tfrecords_reader.read_files([file_instruction], read_config=tfds.ReadConfig(shuffle_seed=seed), shuffle_files=shuffle_files)
        return ds

    def size(self, split):
        """Returns the number of examples in the split."""
        split = self._map_split(split)
        ds_splits = self.info.splits
        dataset_size = ds_splits[split].num_examples
        # Very large datasets have num_examples = 0; default instead to np.inf
        dataset_size = dataset_size if dataset_size > 0 else np.inf
        return dataset_size

def dict_to_tfexample(ex):
    """Convert example dictionary to tf.train.Example proto."""
    feature_dict = {}
    for k, v in ex.items():
        t = tf.constant(v)
        if len(t.shape) == 0:  # pylint:disable=g-explicit-length-test
            v = [v]
        elif len(t.shape) == 1:
            v = list(v)
        else:
            raise ValueError("Unsupported shape (%s) for '%s' value: %s" % (t.shape, k, v))

    if t.dtype == tf.string and len(t.shape) <= 1:
        feature_dict[k] = tf.train.Feature(bytes_list=tf.train.BytesList(value=[tf.compat.as_bytes(t) for t in v]))
    elif t.dtype in (tf.bool, tf.int32, tf.int64) and len(t.shape) <= 1:
        feature_dict[k] = tf.train.Feature(int64_list=tf.train.Int64List(value=v))
    elif t.dtype in (tf.float32, tf.float64) and len(t.shape) <= 1:
        feature_dict[k] = tf.train.Feature(float_list=tf.train.FloatList(value=v))
    else:
        raise ValueError("Unsupported type (%s) and shape (%s) for '%s' value: %s" % (t.dtype, t.shape, k, v))

    return tf.train.Example(features=tf.train.Features(feature=feature_dict))


# =============================== DataSources ==================================


class DataSource(DatasetProviderBase):
    """A `DatasetProvider` that provides raw data from an input source.
    Inherits all abstract methods and properties of `DatasetProviderBase` except
    those overidden below.
    """

    def __init__(self, splits: Iterable[str], num_input_examples: Optional[Mapping[str, int]] = None):
        self._splits = tuple(splits)
        self._num_input_examples = (dict(num_input_examples) if num_input_examples is not None else None)

    @property
    def splits(self) -> Sequence[str]:
        return self._splits

    @property
    def output_features(self) -> Mapping[str, Feature]:
        """Override unused property of `DatasetProviderBase`."""
        raise NotImplementedError

    @abc.abstractmethod
    def list_shards(self, split: str) -> Sequence[str]:
        """Returns string identifiers of input shards."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> tf.data.Dataset:
        """Overrides base class to add shard identifier and remove use_cached.
        Args:
        split: string, the split to return.
        shuffle: bool, whether to shuffle the input source.
        seed: tf.int64 scalar tf.Tensor (or None) for shuffling input source.
        shard_info: optional specification for loading a shard of the split.
        """
        raise NotImplementedError

    def num_input_examples(self, split: str) -> Optional[int]:
        if self._num_input_examples is None:
            return None
        return self._num_input_examples[split]


    def _validate_args(fn, expected_pos_args):
        """Ensure function has exactly expected positional args."""
        argspec = inspect.getfullargspec(fn)
        expected_pos_args = tuple(expected_pos_args)
        actual_args = tuple(argspec.args)
        if actual_args[:len(expected_pos_args)] != expected_pos_args:
            raise ValueError("'%s' must have positional args %s, got: %s" % (fn.__name__, expected_pos_args, actual_args))
        actual_pos_args = tuple(argspec.args[:-len(argspec.defaults)] if argspec.defaults else argspec.args)
        if actual_pos_args != expected_pos_args[:len(actual_pos_args)]:
            raise ValueError("'%s' may only have positional args %s, got: %s" % (fn.__name__, expected_pos_args, actual_pos_args))


class TfdsDataSource(DataSource):
    """A `DataSource` that uses TensorFlow Datasets to provide the input data."""

    def __init__(self, tfds_name: str, tfds_data_dir: Optional[str] = None, splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None):
        """TfdsTask constructor.
        Args:
        tfds_name: string, the name and version number of a TFDS dataset,
            optionally with a config.
        tfds_data_dir: string, an optional path to a specific TFDS data directory
            to use.
        splits: an iterable of allowable string split names, a dict mapping
            allowable canonical splits (e.g., 'validation') to TFDS splits or slices
            (e.g., 'train[':1%']), or None. The default, None, uses all available
            splits from the TFDS dataset info.
        """
        if ":" not in tfds_name:
            raise ValueError("TFDS name must contain a version number, got: %s" % tfds_name)

        if splits and not isinstance(splits, dict):
            splits = {k: k for k in splits}

        self._tfds_dataset = LazyTfdsLoader(tfds_name, data_dir=tfds_data_dir, split_map=splits if isinstance(splits, dict) else None)

        # If splits are not provided, we pass an empty tuple and use the lazy
        # lookup in the `splits` property.
        super().__init__(splits=splits or ())

    @property
    def splits(self):
        """Overrides since we can't call `info.splits` until after init."""
        return self._splits or self._tfds_dataset.info.splits

    @property
    def tfds_dataset(self):
        return self._tfds_dataset

    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> tf.data.Dataset:
        return self.tfds_dataset.load(split, shuffle_files=shuffle, seed=seed, shard_info=shard_info)

    def num_input_examples(self, split: str) -> int:
        """Overrides since we can't call `info.splits` until after init."""
        return self.tfds_dataset.size(split)

    def list_shards(self, split: str) -> Sequence[str]:
        return self.tfds_dataset.files(split)


class T5DataSource(DataSource):
    def __init__(self, config_or_file: Any, preprocessor: Optional[Any] = None, splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None):
        self._tfds_dataset = LazyTFDSIOLoader(config_or_file=config_or_file, preprocessor=preprocessor, split_map=splits if isinstance(splits, dict) else None)
        super().__init__(splits=splits or ())

    @property
    def splits(self):
        """Overrides since we can't call `info.splits` until after init."""
        return self._tfds_dataset.info.splits

    @property
    def tfds_dataset(self):
        return self._tfds_dataset

    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> tf.data.Dataset:
        return self.tfds_dataset.load(split, shuffle_files=shuffle, seed=seed, shard_info=shard_info)

    def num_input_examples(self, split: str) -> int:
        """Overrides since we can't call `info.splits` until after init."""
        return self.tfds_dataset.size(split)

    def list_shards(self, split: str) -> Sequence[str]:
        return self.tfds_dataset.files(split)


def load_sp_model(model_path):
    m = model.ModelProto()
    m.ParseFromString(File.rb(model_path).read())
    return m

def calculate_extra_ids(show_extra=False):
    global _num_extra_ids
    m = load_sp_model(_vocab_path)
    sp_size = len(m.pieces)
    _resize = True
    if sp_size > _default_sp_models['multi']['size']:
        base_size = _default_sp_models['multi']['size']
    elif sp_size > _default_sp_models['default']['size']:
        base_size = _default_sp_models['default']['size']
    else:
        _resize = False
        base_size = sp_size - 100
    if _resize:
        _num_extra_ids = 100 - (sp_size - base_size)
        if show_extra:
            logging.info(f'Showing Extra Tokens')
            for i, piece in enumerate(m.pieces[base_size:]):
                logging.info(f'{i+base_size}: {piece.piece}')

    logging.info(f'Extra IDs: {_num_extra_ids}. SPE Model Size: {sp_size}')


def set_vocab(vocab_path, show_extra=True):
    global _vocab_path
    if _vocab_path and _vocab_path == vocab_path:
        return
    if not File.exists(vocab_path):
        logging.error(f'Error Setting vocab. Path does not exist: {vocab_path}')
        raise ValueError
    _vocab_path = vocab_path
    calculate_extra_ids(show_extra)


def create_sp_model(save_path, special_tokens=None, base_model='default', set_as_global=True, overwrite=False):
    if not save_path.endswith('.model'):
        save_path = File.join(save_path, 'sentencepiece.model')
    if File.exists(save_path) and not overwrite:
        logging.info(f'save_path = {save_path} exists and overwrite = False.')
        if set_as_global:
            logging.info(f'Setting to Global and Returning')
            set_vocab(save_path)
        return
    if base_model not in ['default', 'multi']:
        assert File.exists(base_model)
        m = load_sp_model(base_model)
    else:
        m = load_sp_model(_default_sp_models[base_model]['path'])
    base_size = len(m.pieces)
    logging.info(f'SP Model Size: {base_size}')
    if special_tokens:
        tokens_to_remove = []
        for i, piece in enumerate(m.pieces):
            if str(piece.piece) in special_tokens:
                logging.info(f'{i}: {piece.piece} exists in Vocab')
                tokens_to_remove.append(str(piece.piece))
        if tokens_to_remove:
            special_tokens -= tokens_to_remove
        logging.info(f'Adding {len(special_tokens)} Special Tokens.')
        for token in special_tokens:
            new_token = model.ModelProto().SentencePiece()
            new_token.piece = token
            new_token.score = 0
            m.pieces.append(new_token)
            logging.info(f'Added Token: ID {len(m.pieces)} = {token}')
        logging.info(f'Updated SP Model Size: {len(m.pieces)}')
    with File.wb(save_path) as f:
        f.write(m.SerializeToString())
    if set_as_global:
        logging.info(f'Setting {save_path} to Global Vocab')
        set_vocab(save_path)


def default_output_features(input_key='inputs', target_key='targets', vocab_path=None, show_extra=True):
    if vocab_path:
        set_vocab(vocab_path, show_extra)
    assert _vocab_path is not None
    return {
        input_key: seqio.Feature(
            seqio.SentencePieceVocabulary(_vocab_path, extra_ids=_num_extra_ids),
            add_eos=True, required=False, dtype=tf.int32
        ),
        target_key: seqio.Feature(
            seqio.SentencePieceVocabulary(_vocab_path, extra_ids=_num_extra_ids),
            add_eos=True, dtype=tf.int32
        ),
    }

def t5_preprocessors(input_key='inputs', target_key='targets', is_lm=True, vocab_path=None, show_extra=True):
    if vocab_path:
        set_vocab(vocab_path, show_extra)
    if is_lm:
        input_key = None
    return {
            'unsupervised': [
                functools.partial(
                    t5.data.preprocessors.rekey, 
                    key_map={'inputs': input_key, 'targets': target_key}
                ),
                seqio.preprocessors.tokenize,
                seqio.CacheDatasetPlaceholder(),
                t5.data.preprocessors.unsupervised,
                seqio.preprocessors.append_eos_after_trim
            ],
            'span_corruption': [
                functools.partial(
                    t5.data.preprocessors.rekey, 
                    key_map={'inputs': input_key, 'targets': target_key}
                ),
                seqio.preprocessors.tokenize,
                seqio.CacheDatasetPlaceholder(),
                t5.data.preprocessors.span_corruption,
                seqio.preprocessors.append_eos_after_trim
            ],
            'iid_denoising': [
                functools.partial(
                    t5.data.preprocessors.rekey, 
                    key_map={'inputs': input_key, 'targets': target_key}
                ),
                seqio.preprocessors.tokenize,
                seqio.CacheDatasetPlaceholder(),
                t5.data.preprocessors.iid_denoising,
                seqio.preprocessors.append_eos_after_trim
            ],
            'prefix_lm': [
                functools.partial(
                    t5.data.preprocessors.rekey, 
                    key_map={'inputs': input_key, 'targets': target_key}
                ),
                seqio.preprocessors.tokenize,
                seqio.CacheDatasetPlaceholder(),
                t5.data.preprocessors.prefix_lm,
                seqio.preprocessors.append_eos_after_trim
            ]
        }

def default_preprocessor(*args, **kwargs):
    return [
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        seqio.preprocessors.append_eos_after_trim,
    ]

def default_metrics(*args, **kwargs):
    return [t5.evaluation.metrics.bleu, t5.evaluation.metrics.rouge]

_all_task_names = ['default', 'unsupervised', 'span_corruption', 'iid_denoising', 'prefix_lm']

def add_to_registry(data_config, splits={'train': 'train', 'validation': 'train[0:1000'}, preprocessor=None, tasks=['default'], vocab_path=None):
    global _dataset_registry
    base_name = data_config['name']
    is_lm = bool(data_config['classifier'] == 'lm')
    input_key, target_key = None, None
    for k,v in data_config['datamap'].items():
        if v == 'inputs':
            input_key = k
        elif v == 'targets':
            target_key = k
    logging.info(f'Input Key = {input_key}. Target Key = {target_key}. LM = {is_lm}')
    all_procs = t5_preprocessors(input_key=input_key, target_key=target_key, is_lm=is_lm, vocab_path=vocab_path)
    _default_features = default_output_features(input_key=input_key, target_key=target_key, vocab_path=vocab_path)
    _default_metrics = default_metrics() if not is_lm else []
    _tasks = []
    for task in tasks:
        assert task in _all_task_names, f'Invalid Task: {task}. Valid Tasks: {_all_task_names}'
        task_name = base_name
        if task != 'default':
            task_name += f'_{task}'
        
        if task_name in _dataset_registry:
            logging.error(f'Task: {task_name} already in Dataset Mixtures')
            continue
        
        if task == 'default':
            seqio.TaskRegistry.add(
                task_name,
                source=T5DataSource(config_or_file=data_config, preprocessor=preprocessor, splits=splits),
                preprocessors=default_preprocessor(),
                output_features=_default_features,
                metric_fns=_default_metrics,
            )

        else:
            seqio.TaskRegistry.add(
                task_name,
                source=T5DataSource(config_or_file=data_config, preprocessor=preprocessor, splits=splits),
                preprocessors=all_procs[task],
                output_features=_default_features,
                metric_fns=_default_metrics,
            )
        _tasks.append(task_name)
    logging.info(f'Added all Tasks: {_tasks}')
    _dataset_registry.extend(_tasks)
    
def remove_from_registry(task_names=None, clear_all=False):
    global _dataset_registry
    if not task_names and not clear_all:
        logging.error(f'Nothing to Remove. Current Registry: {_dataset_registry}')
        return
    if task_names:
        for task in task_names:
            if task in _dataset_registry:
                logging.info(f'Removing Task: {task}')
                seqio.MixtureRegistry.remove(task)
                _dataset_registry.remove(task)
            else:
                logging.info(f'Task {task} not found.')
    elif clear_all:
        logging.info(f'Clearing All Tasks from Registry')
        while _dataset_registry:
            task = _dataset_registry.pop(0)
            logging.info(f'Removing Task: {task}')
            seqio.MixtureRegistry.remove(task)

def add_to_mixture(mixture_name, tasks='all', weight_ratio=(3, 1), weighted_task=None):
    global _dataset_mixtures
    if isinstance(tasks, str):
        tasks = _dataset_registry if tasks == 'all' else [tasks]
    _mixtures = []
    for task_name in tasks:
        if task_name not in _dataset_registry:
            logging.info(f'Task: {task_name} not in Registry')
            continue
        task_weight = weight_ratio[0] if (weighted_task and weighted_task in task_name) else weight_ratio[1]
        _mixtures.append((task_name, task_weight))
        logging.info(f'Mixture Task: {task_name} = {task_weight}')
    logging.info(f'Mixtures Added: {_mixtures}')
    _dataset_mixtures[mixture_name] = _mixtures


def remove_from_mixture(mixture_names=None, clear_all=False):
    global _dataset_mixtures
    if not mixture_names and not clear_all:
        logging.error(f'Nothing to Remove. Current Mixtures: {_dataset_mixtures}')
        return
    if mixture_names:
        if isinstance(mixture_names, str):
            mixture_names = [mixture_names]
        for mixture in mixture_names:
            if mixture in _dataset_mixtures:
                logging.info(f'Removing Mixture: {mixture}')
                seqio.MixtureRegistry.remove(mixture)
                _ = _dataset_mixtures.pop(mixture)
            else:
                logging.info(f'Mixture {mixture} not found.')
    elif clear_all:
        logging.info(f'Clearing All Mixtures from _dataset_mixtures')
        all_mixtures = list(_dataset_mixtures.keys())
        for mixture in all_mixtures:
            logging.info(f'Removing Mixture: {mixture}: {_dataset_mixtures[mixture]}')
            _ = _dataset_mixtures.pop(mixture)
            seqio.MixtureRegistry.remove(mixture)


def get_registry():
    logging.info(f'All Registry: {_dataset_registry}')
    return _dataset_registry

def get_mixtures():
    logging.info(f'All Mixtures: {_dataset_mixtures}')
    return _dataset_mixtures

def sample_registry(num_samples=15, split='train', seq_lengths={'inputs': 512, 'targets': 1024}, printer=pprint, return_samples=True):
    printer(f'Sampling Registry')
    enable_v2_behavior()
    _samples = {}
    for task_name in _dataset_registry:
        _samples[task_name] = {}
        printer(f'{num_samples} Examples from {task_name}')
        td = seqio.get_mixture_or_task(task_name).get_dataset(sequence_length=seq_lengths, split=split)
        for x, ex in zip(range(num_samples), td.as_numpy_iterator()):
            printer(ex)
            _samples[task_name][x] = ex
        printer(('----' * 15))
    if return_samples:
        return _samples

def sample_mixture(num_samples=15, split='train', seq_lengths={'inputs': 512, 'targets': 1024}, printer=pprint, return_samples=True):
    printer(f'Sampling Mixtures')
    enable_v2_behavior()
    _samples = {}
    for mixture_name in list(_dataset_mixtures.keys()):
        _samples[mixture_name] = {}
        printer(f'{num_samples} Examples from {mixture_name}')
        td = seqio.get_mixture_or_task(mixture_name).get_dataset(sequence_length=seq_lengths, split=split)
        for x, ex in zip(range(num_samples), td.as_numpy_iterator()):
            printer(ex)
            _samples[mixture_name][x] = ex
        printer(('----' * 15))
    if return_samples:
        return _samples